/*
Very simple GLUT program using bare OpenGL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2006/09/15 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Variables updated by the GUI */
float spin_x, spin_y;

void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATAL OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************ Programmable Shading Support **************/

/**
  Program to run on each vertex or pixel fragment.
*/
class oglShaderObject {
	int target;// Program target, like GL_FRAGMENT_SHADER_ARB
	GLhandleARB progName; // OpenGL handle to program
public:
	oglShaderObject(int target_) :target(target_) 
		{progName=glCreateShaderObjectARB(target);}
	
	/// Get our handle (e.g., for glAttachObjectARB)
	GLhandleARB get(void) const {return progName;}
	
	// Replace the current program by the one in this file.
	//  Prints out an error message if the program can't be parsed.
	void read(const char *fName);
	
	// Replace the current program by this one.
	//  Prints out an error message if the program can't be parsed.
	void set(const char *programText);
	
	~oglShaderObject() {glDeleteObjectARB(progName);}
};

void oglShaderObject::read(const char *fName) {
	
	// Read program from file
	int len=65536;
	char *s=new char[len];
	FILE *f=fopen(fName,"r");
	len=fread(s,1,len,f); // reset len to file length
	s[len]=0; // null terminate string
	fclose(f);
	
	// Upload program to graphics card
	set(s);
	
	delete[] s;
}

void checkOp(GLhandleARB obj,int what,const char *where) {
	GLint compiled;
	glGetObjectParameterivARB(obj,what,&compiled);
	if (!compiled) {
		printf("Compile error on program: %s\n",where);
		enum {logSize=1000};
		char log[logSize]; int len=0;
		glGetInfoLogARB(obj, logSize,&len,log);
		printf("Log: \n%s\n",log);
		exit(1);
	}
}
	
void oglShaderObject::set(const char *programText) {
	glShaderSourceARB(progName,1, &programText,NULL);
	glCompileShaderARB(progName);
	checkOp(progName,GL_OBJECT_COMPILE_STATUS_ARB,programText);
	Check(programText);
}
void uploadShaders(oglShaderObject &vert,oglShaderObject &frag,
	GLhandleARB &prog) 
{
	Check("before uploading shaders");

	vert.read("vertex.txt");
	frag.read("fragment.txt");
	
	glAttachObjectARB(prog,vert.get());
	glAttachObjectARB(prog,frag.get());

	glLinkProgramARB(prog);
	checkOp(prog,GL_OBJECT_LINK_STATUS_ARB,"link");
	Check("link");
	
	if (0) { /* Linker log can show what runs on hardware */
		char log[65536];int loglen=0;
		glGetInfoLogARB(prog,sizeof(log)-1,&loglen,log);
		log[loglen]=0;
		printf("Log: %s\n",log);
	}
}


/************* Object we'll be drawing ***********/
class mainObject_t {
public:
	mainObject_t();
	void draw(void);
};


/** Set up a 3D model, stored in a vertex buffer */
mainObject_t::mainObject_t()
{
	/* nothing to set up here... */
}

/** Main drawing routine, called by main_display below */
void mainObject_t::draw(void)
{
	/* Set up programmable shader.
	  In a real program, you'd do all your glCreateProgramObjectARB 
	  calls at startup, instead of recompiling the program every frame... */
	oglShaderObject vert(GL_VERTEX_SHADER_ARB);   
	oglShaderObject frag(GL_FRAGMENT_SHADER_ARB); 
	
	GLhandleARB prog=glCreateProgramObjectARB();
	uploadShaders(vert,frag,prog);
	
	// Everything we draw from here will use our programmable shader...
	glUseProgramObjectARB(prog);
	Check("use");
	
	/* Copy spin_x and spin_y into program uniforms */
	glUniform1fARB(glGetUniformLocationARB(prog,"spin_x"),spin_x);
	glUniform1fARB(glGetUniformLocationARB(prog,"spin_y"),spin_y);
	
	/* Finally, at long last, draw some friggin' objects! */
	glBegin(GL_TRIANGLES);
	for (float bz=-0.25; bz<=+0.25;bz+=0.01) {
		glColor4f(1,0,0,1); /* Red triangle */
		glVertex4f(0,0,bz,1);
		glVertex4f(1,0,bz,1);
		glVertex4f(0,1,bz,1);

		glColor4f(0,1,0,1); /* Green triangle */
		float z=bz+0.1;
		glVertex4f(1,1,z,1);
		glVertex4f(1,0,z,1);
		glVertex4f(0,1,z,1);
	}
	glEnd();
	
	/* Stop using programmable shader */
	glUseProgramObjectARB(0);
	glDeleteObjectARB(prog);
	
	/* This line causes the world to repeatedly redraw itself, using 100% of the CPU. */
	glutPostRedisplay();
}


/******** Program Initialization *********/
mainObject_t *mainObject=0; /**< Big object we're trying to draw */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Set up a little error-prone OpenGL state:
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw();
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)w, (GLsizei)(h-50)); //<- hack: leave room at top for GLUI spinner boxes
}

int main(int argc, char** argv)
{
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Bare-metal OpenGL Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();

	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	// Make an object to draw
	mainObject=new mainObject_t;
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);
	GLUI_Spinner *spinner_x = new GLUI_Spinner( glui, "X:", &spin_x);
	spinner_x->set_float_limits(0.0, 1.0);
	GLUI_Spinner *spinner_y = new GLUI_Spinner( glui, "Y:", &spin_y);
	spinner_y->set_float_limits(0.0, 1.0);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
